""" module to replace PyGreSQL _pg.so

    LICENSE
    =======

    Copyright (c) 2004, Randall Smith
    All rights reserved.

    Redistribution and use in source and binary forms, with or without 
    modification, are permitted provided that the following conditions are met:

        * Redistributions of source code must retain the above copyright 
          notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright 
          notice, this list of conditions and the following disclaimer in the 
          documentation and/or other materials provided with the distribution.
        * Neither the name of the <ORGANIZATION> nor the names of its 
          contributors may be used to endorse or promote products derived from 
          this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
    CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
    SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
    INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
    CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
    ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
    POSSIBILITY OF SUCH DAMAGE.

    NOTES
    =======
    This is alpha software.  Don't use it on valuable data.

    It replaces the _pg.so module included with PyGreSQL to make this a 100%
    pure Python database module.  I created as an experiment to determine the
    feasability of creating pure Python drivers for databases.

"""

import os
import socket
import pgmessages
from PGSocket import PGSocket

RESULT_DQL = 4

class Connection(object):

    def __init__(self, host, port, dbname, user, passwd):
        # If we are missing a username, try to get it.
        if user is None:
            try:
                user = os.getlogin()
            except:
                raise Exception, 'user is required.'
        if dbname is None:
            dbname = user
        self.host = host
        self.port = port
        self.dbname = dbname
        self.user = user
        self.passwd = passwd
        self.params = {}
        self.connect()

    def connect(self):
        # If host is provided and Unix sockets available, do so.
        if self.host is not None:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((self.host, self.port))
        elif self.host is None and socket.AF_UNIX:
            try:
                sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                sock.connect('/var/run/postgresql/.s.PGSQL.5432')
            except:
                self.host = 'localhost'
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        else: 
            self.host = 'localhost'
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((self.host, self.port))
        # Set timeout so we don't hang forever if something goes wrong.
        sock.settimeout(10)
        self._sock = sock
        startup_message = pgmessages.StartupMessage(self.user, self.dbname)
        sock.send(startup_message.server_message)
        # Now that the startup message is sent, use our custom socket.
        pgsock = PGSocket(sock)
        self._pgsock = pgsock
        message = pgsock.getMessage()
        if message.format_name == 'ErrorResponse':
            raise Exception, 'got an error.  Probably IDENT failure'
        if isinstance(message, pgmessages.AuthRequestMessage):
            password_type = message.specifier
            salt = message.salt
            msg = pgmessages.PasswordMessage(password_type, self.passwd,
                                             self.user, salt)
            pgsock.send(msg.server_message)
        nextmsg = pgsock.getMessage()
        if nextmsg.format_name == 'ErrorResponse':
            raise Exception, 'Authentication Failed.'
        elif nextmsg.format_name == 'AuthenticationOk':
            pass
        elif nextmsg.format_name == 'ParameterStatus':
            # Must have used ident.
            pass
        else:
            emsg = 'Did not handle response, %s' % nextmsg.format_name
            raise Exception, emsg
        while nextmsg.format_name != 'ReadyForQuery':
            if nextmsg.format_name == 'ParameterStatus':
                self.params[nextmsg.param_name] = nextmsg.param_value
            nextmsg = pgsock.getMessage()

    def source(self):
        return Source(self._pgsock)

    def close(self):
        self._pgsock.send(pgmessages.Terminate().server_message)
        self._sock.close()

class Source(object):
    def __init__(self, pgsock):
        self.pgsock = pgsock
        self.resulttype = 0
        self.ntuples = 0
        self.column_info = None
        self.__resultpos = 0

    def execute(self, query):
        """Execute a simple query.
        
        Return the number of rows affected or None for a query that actually
        returns rows.        
        """
        pgsock = self.pgsock
        querymsg = pgmessages.Query(query).server_message
        pgsock.send(querymsg)
        results = []
        nextmsg = pgsock.getMessage()
        if nextmsg.format_name == 'RowDescription':
            self.column_info = {}
            self.column_info['count'] = nextmsg.colnum
            self.column_info['attributes'] = nextmsg.columns
        ##print nextmsg.format_name
        while nextmsg.format_name <> 'ReadyForQuery':
            nextmsg = pgsock.getMessage()
            if nextmsg.format_name == 'DataRow':
                results.append(nextmsg.rowdata)
        # Fix this later.
        self._rset = results
        self.ntuples = len(results)
        self.resulttype = 4
        self.__resultpos = 0
        return None

    def listinfo(self):
        """
		listinfo is a sequence of
		(index, column_name, type_oid)
		getdescr returns all items needed for a
		description tuple except the column_name.

        Appears to be a tuple of tuples.  The index is just an incrementing
        number starting with zero.
        """
        if self.column_info is not None:
            info = []
            columns = self.column_info['attributes']
            column_count = len(columns)
            for i in range(column_count):
                column = columns[i]
                column_info = (i, column['name'], column['type_oid'])
                info.append(column_info)
            return tuple(info)
        else:
            return ()

    def fetch(self, numrows):
        """Return results.
       
        -1 All Rows
        """
        if numrows == -1:
            subset = self._rset[self.__resultpos:]
            self.__resultpos = self.ntuples
        elif numrows > self.ntuples - self.__resultpos:
            subset = self._rset[self.__resultpos:]
            self.__resultpos = self.ntuples
        else:
            newpos = self.__resultpos + numrows
            subset = self._rset[self.__resultpos:newpos]
            self.__result_pos = newpos
        return subset
		
def connect(dbbase, dbhost, dbport, dbpot, dbtty, dbuser, dbpasswd):
    if dbport == -1:
        dbport = 5432
    return Connection(dbhost, dbport, dbbase, dbuser, dbpasswd)
   

